from .add import Add
from .add_constant import AddConstant
from .average_pooling_2d import AveragePooling2D
from .batch_normalization import BatchNormalization
from .bias import Bias
from .bilinear_2d import Bilinear2D
from .broadcast import BroadcastTo
from .clipped_relu import ClippedReLU
from .concat import Concat
from .constant import Constant
from .convolution_2d import Convolution2D
from .cos import Cos
from .depthwise_convolution_2d import DepthwiseConvolution2D
from .div import Div
from .dropout import Dropout
from .elu import ELU
from .exp import Exp
from .function import Function
from .gemm import Gemm
from .leaky_relu import LeakyReLU
from .linear import Linear
from .local_response_normalization import LocalResponseNormalization
from .matmul import MatMul
from .max_pooling_2d import MaxPooling2D
from .mul import Mul
from .mul_constant import MulConstant
from .pad import ConstantPadding
from .relu import ReLU
from .reshape import Reshape
from .resize_2d import Resize2D
from .scale import Scale
from .sigmoid import Sigmoid
from .sin import Sin
from .slice import Slice
from .softmax import Softmax
from .softmax_cross_entropy import SoftmaxCrossEntropy
from .sub import Sub
from .sum import Sum
from .swish import Swish
from .tan import Tan
from .tanh import Tanh
from .transpose import Transpose
from .unpooling_2d import Unpooling2D



import argparse
import io
import sys
from typing import Any, Dict, Iterator, List

import msgpack
import numpy
from numpy.typing import NDArray


def function_label(function: Dict[bytes, Any]) -> str:
    ret = "{{" + (
        "|".join(
            [function[b"name"].decode()]
            + list(map(lambda v: "<a" + str(v[0]) + v[1].decode() + ">", enumerate(function[b"inputs"])))
        )
    )
    if b"W" in function[b"params"]:
        ret += "|W"
    if b"b" in function[b"params"]:
        ret += "|b"
    ret += "}"

    def find_params(params: Dict[bytes, Any]) -> Iterator[str]:
        for k, v in params.items():
            if type(v) is dict and b"ndarray" in v:
                array = numpy.load(io.BytesIO(v[b"ndarray"]))  # type: ignore
                yield k.decode() + " shape: " + str(array.shape)
            elif type(v) is dict and b"nnoir" in v:
                yield k.decode() + ": " + v[b"nnoir"][b"model"][b"name"].decode("utf-8")
            else:
                yield k.decode() + ": " + str(v)

    params_str = "&#92;l".join(find_params(function[b"params"]))
    if params_str != "":
        ret += "|{%s&#92;l}" % params_str
    ret += "|{%s}}" % "|".join(map(lambda v: "<" + v.decode() + ">", reversed(function[b"outputs"])))
    return ret


def function_name(function: Dict[bytes, Any]) -> str:
    inputs = "".join(map(lambda v: v.decode(), reversed(function[b"inputs"])))
    outputs = "".join(map(lambda v: v.decode(), reversed(function[b"outputs"])))
    return "{}_{}_{}".format(function[b"name"].decode(), inputs, outputs)


def value_name(value: Dict[bytes, Any]) -> str:
    return "{} {}".format(value[b"name"].decode(), tuple(value[b"shape"]))


def to_dot(nnoir: Dict[bytes, Any], rankdir: str = "TB") -> str:
    inputs = nnoir[b"nnoir"][b"model"][b"inputs"]
    outputs = nnoir[b"nnoir"][b"model"][b"outputs"]
    values = nnoir[b"nnoir"][b"model"][b"values"]
    functions = nnoir[b"nnoir"][b"model"][b"functions"]

    ret = "digraph graphname { rankdir=%s;\n" % rankdir
    ret += "  subgraph input {\n"
    for i, var in enumerate(inputs):
        attribute = {"xlabel": "input%d = %s" % (i, var.decode()), "shape": "point"}
        attributes = ['%s="%s"' % (k, v) for (k, v) in attribute.items()]
        ret += "    %s [%s];\n" % (var.decode(), ",".join(attributes))
    ret += "  }\n"
    ret += "  subgraph output {\n"
    for i, var in enumerate(outputs):
        attribute = {"xlabel": "output%d = %s" % (i, var.decode()), "shape": "point"}
        attributes = ['%s="%s"' % (k, v) for (k, v) in attribute.items()]
        ret += "    %s [%s];\n" % (var.decode(), ",".join(attributes))
    ret += "  }\n"
    for value in values:
        if value[b"name"] in inputs:
            continue
        if value[b"name"] in outputs:
            continue
        attribute = {"xlabel": value_name(value), "shape": "point"}
        attributes = ['%s="%s"' % (k, v) for (k, v) in attribute.items()]
        ret += "  %s [%s];\n" % (value[b"name"].decode(), ",".join(attributes))
    for function in functions:
        # output function
        attribute = {"label": function_label(function), "shape": "record", "style": "filled", "fillcolor": "aquamarine"}
        attributes = ['%s="%s"' % (k, v) for (k, v) in attribute.items()]
        ret += "  %s [%s];\n" % (function_name(function), ",".join(attributes))
        # output functions
        for i, from_value in enumerate(function[b"inputs"]):
            ret += "  %s -> %s:a%d%s[constraint=false];" % (
                from_value.decode(),
                function_name(function),
                i,
                from_value.decode(),
            )
            ret += '  %s -> %s[color="transparent"];\n' % (from_value.decode(), function_name(function))
        for to_value in function[b"outputs"]:
            ret += "  %s:%s -> %s[constraint=false];" % (function_name(function), to_value.decode(), to_value.decode())
            ret += '  %s -> %s[color="transparent"];\n' % (function_name(function), to_value.decode())

    ret += "  { rank = same; \n"
    for var in inputs:
        attribute = {"xlabel": var.decode(), "shape": "point"}
        attributes = ['%s="%s"' % (k, v) for (k, v) in attribute.items()]
        ret += "    %s;\n" % (var.decode())
    ret += "  }\n"
    ret += "  { rank = same; \n"
    for var in outputs:
        attribute = {"xlabel": var.decode(), "shape": "point"}
        attributes = ['%s="%s"' % (k, v) for (k, v) in attribute.items()]
        ret += "    %s;\n" % (var.decode())
    ret += "  }\n"
    ret += "}"
    return ret


def nnoir2dot() -> None:
    parser = argparse.ArgumentParser(description="NNOIR to Graphviz(dot) Converter")
    parser.add_argument(dest="input", type=str, metavar="NNOIR", help="input(NNOIR) file path")
    args = parser.parse_args()
    with open(args.input, "rb") as f:
        print(to_dot(msgpack.unpackb(f.read(), raw=True)))



import argparse
import sys
from typing import List

import numpy as np
from nnoir import NNTC, TestValue, Value, load
from numpy.typing import NDArray


def pack_values(ns: List[bytes], vs: List[NDArray[np.float32]]) -> List[TestValue]:
    return [TestValue(n, v) for (n, v) in zip(ns, vs)]


def generate_inputs(vs: List[Value], seed: int, base: float, scale: float) -> List[NDArray[np.float32]]:
    inputs = []
    np.random.seed(seed)
    for v in vs:
        x = base + scale * np.random.randn(*v.shape).astype(np.float32)  # type: ignore
        inputs.append(x)
    return inputs


def main() -> None:
    parser = argparse.ArgumentParser(description="NNTC generator")
    parser.add_argument(dest="nnoir", type=str, metavar="NNOIR", help="input nnoir file")
    parser.add_argument("-o", "--output", dest="output", type=str, required=True, metavar="PATH", help="output file path")
    parser.add_argument("--seed", dest="seed", type=int, metavar="INT", help="seed for inputs generation", default=0)
    parser.add_argument("--base", dest="base", type=float, metavar="FLOAT", help="modify randn", default=0.0)
    parser.add_argument("--scale", dest="scale", type=float, metavar="FLOAT", help="modify randn", default=1.0)

    args = parser.parse_args()

    nnoir_model = load(args.nnoir)
    if nnoir_model.name == b"main":
        sys.exit(f"Error: invalid model name ({nnoir_model.name.decode()})")

    input_values = [[v for v in nnoir_model.values if v.name == n][0] for n in nnoir_model.inputs]

    inputs = generate_inputs(input_values, args.seed, args.base, args.scale)
    outputs = nnoir_model.run(*inputs)

    tc = NNTC(nnoir_model.name, pack_values(nnoir_model.inputs, inputs), pack_values(nnoir_model.outputs, outputs))
    tc.dump(args.output)



from typing import Any, List, Set

import numpy as np

from . import util
from .function import Function


class Slice(Function):
    def __init__(self, inputs: List[bytes], outputs: List[bytes], **params: Any):
        required_params = {"starts", "ends"}
        optional_params: Set[str] = set()
        super(Slice, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):  # type: ignore
        starts = np.array(self.params["starts"])
        ends = np.array(self.params["ends"])

        indices = [np.s_[starts[i] : ends[i]] for i in range(np.ndim(x))]
        R = x[tuple(indices)].copy()
        return R



import io
from typing import Any, Dict, List

import msgpack
import numpy as np
from numpy.typing import NDArray


class TestValue:
    def __init__(self, name: bytes, ndarray: NDArray[Any]):
        self.name = name
        self.ndarray = ndarray

    def to_dict(self) -> Dict[bytes, Any]:
        def encode_ndarray(obj: NDArray[Any]) -> bytes:
            x = None
            with io.BytesIO() as out:
                np.save(out, obj.copy())  # type: ignore
                x = out.getvalue()
            return x

        return {b"value_name": self.name, b"ndarray": encode_ndarray(self.ndarray)}


class TestCase:
    def __init__(self, inputs: List[TestValue], outputs: List[TestValue]):
        self.inputs = inputs
        self.outputs = outputs

    def to_dict(self) -> Dict[bytes, Any]:
        return {
            b"inputs": [i.to_dict() for i in self.inputs],
            b"outputs": [o.to_dict() for o in self.outputs],
        }


class NNTC:
    def __init__(self, model_name: bytes, inputs: List[TestValue], outputs: List[TestValue]):
        self.version = 0
        self.model = model_name
        self.test_case = TestCase(inputs, outputs)

    def to_dict(self) -> Dict[bytes, Any]:
        return {
            b"version": self.version,
            b"model": self.model,
            b"test_case": self.test_case.to_dict(),
        }

    def pack(self) -> bytes:
        return msgpack.packb(self.to_dict(), use_bin_type=False)  # type: ignore

    def dump(self, file_name: str) -> None:
        result = self.pack()
        with open(file_name, "w") as f:
            f.buffer.write(result)


def _load_test_value(v: Dict[bytes, Any]) -> TestValue:
    name = v[b"value_name"]
    arr = np.load(io.BytesIO(v[b"ndarray"]))  # type: ignore
    return TestValue(name, arr)


def load_nntc(file_name: str) -> NNTC:
    with open(file_name, "rb") as f:
        nntc = msgpack.unpackb(f.read(), raw=True)
    version = nntc[b"version"]
    model = nntc[b"model"]
    inputs = [_load_test_value(v) for v in nntc[b"test_case"][b"inputs"]]
    outputs = [_load_test_value(v) for v in nntc[b"test_case"][b"outputs"]]
    assert version == 0
    return NNTC(model, inputs, outputs)



from typing import Any, List, Set

import numpy as np

from .function import Function


class Cos(Function):
    def __init__(self, inputs: List[bytes], outputs: List[bytes], **params: Any):
        required_params: Set[str] = set()
        optional_params: Set[str] = set()
        super(Cos, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):  # type: ignore
        R = x.copy()
        return np.cos(R)



import argparse

from .load import load
from .nnoir import NNOIR


def read_name(nnoir: NNOIR) -> str:
    return nnoir.name.decode("utf-8")


def read_description(nnoir: NNOIR) -> str:
    return nnoir.description.decode("utf-8")


def read_generator_name(nnoir: NNOIR) -> str:
    return nnoir.generator_name.decode("utf-8")


def read_generator_version(nnoir: NNOIR) -> str:
    return nnoir.generator_version.decode("utf-8")


def write_name(nnoir: NNOIR, name: str) -> None:
    nnoir.name = name.encode(encoding="utf-8")


def write_description(nnoir: NNOIR, description: str) -> None:
    nnoir.description = description.encode(encoding="utf-8")


def print_metadata(nnoir: NNOIR) -> None:
    print(f"name = {read_name(nnoir)}")
    print(f"description = {read_description(nnoir)}")
    print(f"generator.name = {read_generator_name(nnoir)}")
    print(f"generator.version = {read_generator_version(nnoir)}")


def nnoir_metadata() -> None:
    parser = argparse.ArgumentParser(description="read/write NNOIR meta data")
    parser.add_argument(dest="input", type=str, metavar="NNOIR", help="input(NNOIR) file path")
    parser.add_argument(
        "--write-description", type=str, dest="new_description", metavar="<new description>", help="overwrite description"
    )
    parser.add_argument("--write-name", type=str, dest="new_name", metavar="<new name>", help="overwrite name")

    args = parser.parse_args()
    nnoir = load(args.input)

    if args.new_description is None and args.new_name is None:
        return print_metadata(nnoir)

    if args.new_description is not None:
        write_description(nnoir, args.new_description)
    if args.new_name is not None:
        write_name(nnoir, args.new_name)

    nnoir.dump(args.input)



from typing import Any, Dict, Optional, Tuple

from numpy.typing import DTypeLike, NDArray


class Value:
    def __init__(
        self,
        name: bytes,
        np_array: Optional[NDArray[Any]] = None,
        dtype: Optional[bytes] = None,
        shape: Optional[Tuple[int, ...]] = None,
    ):
        self.name = name
        if np_array is None:
            self.dtype = dtype
            self.shape = shape
        else:
            self.dtype = np_array.dtype.str.encode()
            self.shape = np_array.shape

    def dump(self) -> Dict[bytes, Any]:
        return {b"name": self.name, b"dtype": self.dtype, b"shape": self.shape}



from typing import Any, List, Set

import numpy as np

from .function import Function


class Resize2D(Function):
    def __init__(self, inputs: List[bytes], outputs: List[bytes], **params: Any):
        required_params = {"size", "interpolation_mode", "coordinate_transformation_mode"}
        optional_params: Set[str] = set()
        super(Resize2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):  # type: ignore
        out_h = self.params["size"][0]
        out_w = self.params["size"][1]
        modes = self.params["interpolation_mode"].decode("utf-8").split("-")
        if modes[0] == "linear":
            return _run_linear(x, out_h, out_w, self.params["coordinate_transformation_mode"])  # type: ignore
        elif modes[0] == "nearest":
            return _run_nearest(x, out_h, out_w, self.params["coordinate_transformation_mode"], modes[1])  # type: ignore
        else:
            raise Exception("unknow Resize mode")


def _run_linear(x, out_h, out_w, coordinate_transformation_mode):  # type: ignore
    batch, ch, in_h, in_w = x.shape
    if coordinate_transformation_mode == b"align_none":
        # Mode: TF1(default), onnxruntime
        h_scale = float(in_h) / float(out_h)
        w_scale = float(in_w) / float(out_w)
        V = np.minimum(np.arange(out_h) * h_scale, in_h - 1).reshape(out_h, 1)
        V0 = np.maximum(0, np.minimum(np.floor(V), in_h - 2)).astype("int32")
        U = np.minimum(np.arange(out_w) * w_scale, in_w - 1).reshape(1, out_w)
        U0 = np.maximum(0, np.minimum(np.floor(U), in_w - 2)).astype("int32")
    elif coordinate_transformation_mode == b"align_corners":
        # Mode: TF1(align_corners), Chainer, PyTorch
        h_scale = float(in_h - 1) / float(out_h - 1)
        w_scale = float(in_w - 1) / float(out_w - 1)
        V = (np.arange(out_h) * h_scale).reshape(out_h, 1)
        V0 = np.maximum(0, np.minimum(np.floor(V), in_h - 2)).astype("int32")
        U = (np.arange(out_w) * w_scale).reshape(1, out_w)
        U0 = np.maximum(0, np.minimum(np.floor(U), in_w - 2)).astype("int32")
    elif coordinate_transformation_mode == b"align_centers":
        # Mode: TF2, OpenCV
        h_scale = float(in_h) / float(out_h)
        w_scale = float(in_w) / float(out_w)
        V = np.minimum(
            np.maximum((np.arange(out_h) + 0.5) * h_scale - 0.5, 0.0),
            float(in_h - 1),
        ).reshape(out_h, 1)
        V0 = np.maximum(0, np.minimum(np.floor(V), in_h - 2)).astype("int32")
        U = np.minimum(
            np.maximum((np.arange(out_w) + 0.5) * w_scale - 0.5, 0.0),
            float(in_w - 1),
        ).reshape(1, out_w)
        U0 = np.maximum(0, np.minimum(np.floor(U), in_w - 2)).astype("int32")
    else:
        raise Exception("unknown Bilinear mode")
    V1 = V0 + 1
    U1 = U0 + 1
    W00 = (U1 - U) * (V1 - V)
    W01 = (U - U0) * (V1 - V)
    W10 = (U1 - U) * (V - V0)
    W11 = (U - U0) * (V - V0)
    V0 = V0.reshape(out_h)
    V1 = np.minimum(V1.reshape(out_h), in_h - 1)
    U0 = U0.reshape(out_w)
    U1 = np.minimum(U1.reshape(out_w), in_w - 1)
    x0 = x[:, :, V0, :]
    x1 = x[:, :, V1, :]
    result = W00 * x0[:, :, :, U0] + W01 * x0[:, :, :, U1] + W10 * x1[:, :, :, U0] + W11 * x1[:, :, :, U1]
    return result


def _round_index(i, nearest_mode):  # type: ignore
    if nearest_mode == "floor":
        return int(np.floor(i))
    else:
        raise Exception("unsupported nearest_mode")


def _get_original_index(i, l0, l1, coordinate_transformation_mode, nearest_mode):  # type: ignore
    scale = float(l0) / float(l1)
    if coordinate_transformation_mode == b"asymmetric":
        i_ori = i * scale
    else:
        raise Exception("unsupported coordinate_transformation_mode")
    return _round_index(i_ori, nearest_mode)  # type: ignore


def _run_nearest(x, out_h, out_w, coordinate_transformation_mode, nearest_mode):  # type: ignore
    batch, ch, in_h, in_w = x.shape
    y = np.empty((batch, ch, out_h, out_w), dtype=x.dtype)
    for b in np.arange(batch):
        for c in np.arange(ch):
            for i1 in np.arange(out_h):
                i0 = _get_original_index(i1, in_h, out_h, coordinate_transformation_mode, nearest_mode)  # type: ignore
                for j1 in np.arange(out_w):
                    j0 = _get_original_index(j1, in_w, out_w, coordinate_transformation_mode, nearest_mode)  # type: ignore
                    y[b, c, i1, j1] = x[b, c, i0, j0]
    return y



from typing import Any, Dict, List, Set, Tuple

import numpy as np
from numpy.typing import NDArray

from .function import Function


class Add(Function):
    def __init__(self, inputs: List[bytes], outputs: List[bytes], **params: Any):
        required_params: Set[str] = set()
        optional_params: Set[str] = set()
        super(Add, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x1, x2):  # type: ignore
        return x1 + x2
